"""Top-level package for Kleingrid."""

__author__ = """Fabien Mathieu"""
__email__ = 'loufab@gmail.com'
__version__ = '0.1.0'

from kleinberg_grid_simulator.python_implementation.shortcuts import radius2shortcut
from kleinberg_grid_simulator.python_implementation.python_edt import python_edt
from kleinberg_grid_simulator.julia_implementation.julia_edt import julia_edt
from kleinberg_grid_simulator.kleingrid import compute_edt, parallelize, estimate_alpha, get_bounds
